#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import threading
from collections import deque
from datetime import datetime
from functools import total_ordering, wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Union,
    cast,
)

import pendulum
from loguru import logger
from omegaconf import DictConfig
from pendulum.datetime import DateTime
from pendulum.duration import Duration

from abses._bases.components import _Component
from abses._bases.logging import log_session

if TYPE_CHECKING:
    from .main import MainModel

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

VALID_DT_ATTRS = (
    "years",
    "months",
    "weeks",
    "days",
    "hours",
    "minutes",
    "seconds",
)


def time_condition(condition: dict, when_run: bool = True) -> Callable:
    """
    A decorator to run a method based on a time condition.

    Parameters:
        condition:
            A dictionary containing conditions to check against the `time` attribute.
            The keys can be ['year', 'month', 'weekday', 'freqstr'].
        when_run:
            If True, the decorated method will run when the condition is met.
            If False, the decorated method will not run when the condition is met.

    Example:
        ```
        class TestActor(Actor):
            @time_condition(condition={"month": 1, "day": 1}, when_run=True)
            def happy_new_year(self):
                print("Today is 1th, January, Happy new year!")


        parameters = {"time": {"start": "1996-12-24", "days": 1}}


        model = MainModel(parameters=parameters)
        agent = model.agents.new(TestActor, 1, singleton=True)

        for _ in range(10):
            print(f"Time now is {model.time}")
            model.time.go()
            agent.happy_new_year()
        ```
        It should be called again in the next year beginning (i.e., `1998-01-01`) if we run this model longer... It means, the function will be called when the condition is fully satisfied.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not hasattr(self, "time"):
                raise AttributeError(
                    "The object doesn't have a TimeDriver object as `time` attribute."
                )
            time = self.time
            if not isinstance(time, TimeDriver):
                raise TypeError("The `TimeDriver` must be existing.")

            satisfied = all(
                getattr(time.dt, key, None) == value
                for key, value in condition.items()
            )

            if (satisfied and when_run) or (not satisfied and not when_run):
                return func(self, *args, **kwargs)

        return wrapper

    return decorator


@total_ordering
class TimeDriver(_Component):
    """TimeDriver provides the functionality to manage time.

    This class is responsible for managing the time of a simulation model. It keeps track of the current time period, updates it according to a given frequency, and provides properties to access different components of the current time period (e.g., day, hour, etc.). The `TimeDriver` class is a singleton, meaning that there can be only one instance of it per simulation model.

    When init a `TimeDriver`, it accepts below parameters:

    | Parameter Name | Expected Data Type | Default Value | Description |
    |----------------|--------------------|---------------|-------------|
    | start          | str, None                | None          | If None: use the current time, else: should be a string which can be parsed by `pendulum.parse()`. |
    | end            | str, int, None         | None          | If it's a string that can be parsed into datetime the model should end until achieving this time; if int: the model should end in that tick; if None no auto-end. |
    | irregular         | bool               | False         | If False: not dive into an irregular mode (tick-mode); if True, the model will solve as an irregular mode. |
    | years          | int                | 0             | Time duration in years for the duration mode. |
    | months         | int                | 0             | Time duration in months for the duration mode. |
    | weeks          | int                | 0             | Time duration in weeks for the duration mode. |
    | days           | int                | 0             | Time duration in days for the duration mode. |
    | hours          | int                | 0             | Time duration in hours for the duration mode. |
    | minutes        | int                | 0             | Time duration in minutes for the duration mode. |
    | seconds        | int                | 0             | Time duration in seconds for the duration mode. |

    See tutorial to see more.
    """

    _instances: Dict[MainModel[Any, Any], Self] = {}
    _lock = threading.Lock()

    def __new__(cls, model: MainModel):
        with cls._lock:
            if not cls._instances.get(model):
                driver = super(TimeDriver, cls).__new__(cls)
                cls._instances[model] = driver
            else:
                driver = cls._instances[model]
        return driver

    def __init__(self, model: MainModel):
        super().__init__(model=model, name="time")
        self._model: MainModel[Any, Any] = model
        self._tick: int = 0
        self._start_dt: DateTime = pendulum.instance(datetime.now(), tz=None)
        self._end_dt: Optional[Union[DateTime, int]] = None
        self._duration: Optional[Duration] = None
        self._history: deque = deque()
        self._parse_time_settings()
        self._logging_setup()

    def __repr__(self) -> str:
        if self.ticking_mode == "tick":
            return f"<TimeDriver: tick[{self.tick}]>"
        if self.ticking_mode == "duration":
            return f"<TimeDriver: {self.strftime('%Y-%m-%d %H:%M:%S')}>"
        return f"<TimeDriver: irregular[{self.tick}] {self.strftime('%Y-%m-%d %H:%M:%S')}>"

    def __eq__(self, other: object) -> bool:
        return self.dt.__eq__(other)

    def __lt__(self, other: object) -> bool:
        if isinstance(other, (DateTime, datetime)):
            return self.dt < other
        raise NotImplementedError(f"Cannot compare with {other}.")

    @property
    def fmt(self) -> str:
        """String format of datetime."""
        hms = ("hours", "minutes", "seconds")
        if any(getattr(self.duration, item, None) for item in hms):
            return r"%Y-%m-%d %H:%M:%S"
        return r"%Y-%m-%d"

    @property
    def expected_ticks(self) -> Optional[int]:
        """Returns the expected ticks."""
        if not self.end_dt:
            return None
        if isinstance(self.end_dt, datetime):
            if self.duration is None:
                raise RuntimeError("No duration settings.")
            duration = self.dt.diff(self.end_dt)
            steps = duration.total_seconds() / self.duration.total_seconds()
            return int(steps)
        return self.end_dt

    @property
    def should_end(self) -> bool:
        """Should the model end or not."""
        if not self.end_dt:
            return False
        if isinstance(self.end_dt, datetime):
            return self.dt >= self.end_dt
        return self._tick >= self.end_dt

    @property
    def tick(self) -> int:
        """Returns the current tick."""
        return self._tick

    @property
    def ticking_mode(self) -> str:
        """Returns the ticking mode of the time driver."""
        if self.duration:
            return "duration"
        return "irregular" if self._irregular else "tick"

    @property
    def history(self) -> List[datetime]:
        """Returns the history of the time driver."""
        return list(self._history)

    def go(self, ticks: int = 1, **kwargs) -> None:
        """Increments the tick.

        Parameters:
            ticks:
                How many ticks to increase.
        """
        if ticks < 0:
            raise ValueError("Ticks cannot be negative.")
        if ticks == 0 and self.ticking_mode != "irregular":
            raise ValueError(
                "Ticks cannot be zero unless the ticking mode is 'irregular'."
            )
        if ticks > 1:
            for _ in range(ticks):
                self.go(ticks=1, **kwargs)
            return
        # tick = 1
        dt_msg = f" {self.strftime()}" if self.duration else ""
        tick_msg = f" [tick {self.tick}] "
        logger.bind(no_format=True).info(
            "\n" + f"{dt_msg}{tick_msg}".center(30, "-")
        )
        self._tick += ticks
        if self.ticking_mode == "duration":
            self.dt += self.duration
            self._history.append(self.dt)
        elif self.ticking_mode == "irregular":
            self.dt += pendulum.duration(**kwargs)
            self._history.append(self.dt)
        elif self.ticking_mode != "tick":
            raise ValueError(f"Invalid ticking mode: {self.ticking_mode}")
        # end going
        if self.should_end:
            self._model.running = False

    def _parse_time_settings(self) -> None:
        """Setup the time driver."""
        # Parse the start time settings
        self.start_dt = self.params.get("start", None)

        # Parse the end time settings
        self.end_dt = self.params.get("end", None)

        # Parse the duration settings
        self.parse_duration(self.params)

        # Parse the irregular settings
        self.irregular: bool = self.params.get("irregular", False)

        self.dt = self.start_dt
        self._history.append(self.dt)

    def _logging_setup(self) -> None:
        if self.ticking_mode == "duration":
            end = (
                self.end_dt.strftime(self.fmt)
                if isinstance(self.end_dt, datetime)
                else str(self.end_dt)
            )
            msg = (
                f"Ticking mode: {self.ticking_mode}\n"
                f"Start time: {self.start_dt.strftime(self.fmt)}\n"
                f"End time: {end}\n"
                f"Duration: {self.duration}\n"
            )
        else:
            msg = (
                f"Ticking mode: {self.ticking_mode}\n"
                f"Start time: {self.start_dt.strftime(self.fmt)}\n"
                f"End time: {self.end_dt}\n"
            )
        log_session(title="TimeDriver", msg=msg)

    @property
    def irregular(self) -> bool:
        """Returns the irregular mode of the time driver."""
        return self._irregular

    @irregular.setter
    def irregular(self, value: bool) -> None:
        """Set the irregular mode of the time driver."""
        if not isinstance(value, bool):
            raise TypeError("Irregular mode must be a boolean.")
        self._irregular = value

    @property
    def duration(self) -> Duration | None:
        """Returns the duration of the time driver."""
        return self._duration

    def parse_duration(self, duration: DictConfig) -> None:
        """Set the duration of the time driver."""
        valid_attributes = VALID_DT_ATTRS
        valid_dict = {}
        for attribute in valid_attributes:
            value = duration.get(attribute, 0)
            if not isinstance(value, int):
                raise TypeError(f"{attribute} must be an integer.")
            valid_dict[attribute] = value
        if all(value == 0 for value in valid_dict.values()):
            self._duration = None
        else:
            self._duration = pendulum.duration(**valid_dict)

    @property
    def start_dt(self) -> DateTime:
        """Returns the starting time for the model."""
        return self._start_dt

    @start_dt.setter
    def start_dt(self, dt: Optional[Union[datetime, DateTime, str]]) -> None:
        """Set the starting time."""
        if isinstance(dt, datetime):
            self._start_dt = pendulum.instance(dt, tz=None)
        elif dt is None:
            self._start_dt = pendulum.instance(datetime.now(), tz=None)
        elif isinstance(dt, str):
            self._start_dt = cast(DateTime, pendulum.parse(dt, tz=None))
        else:
            raise TypeError(
                "Start time must be a datetime object or a string."
            )

    @property
    def end_dt(self) -> Optional[Union[datetime, int]]:
        """The real-world time or the ticks when the model should be end."""
        return self._end_dt

    @end_dt.setter
    def end_dt(self, dt: Optional[Union[datetime, DateTime, str]]) -> None:
        if isinstance(dt, int) and dt <= 0:
            raise ValueError("End time cannot be negative.")
        if isinstance(dt, int):
            self._end_dt = dt
        elif isinstance(dt, datetime):
            self._end_dt = pendulum.instance(dt, tz=None)
        elif dt is None:
            self._end_dt = None
        elif isinstance(dt, str):
            self._end_dt = cast(DateTime, pendulum.parse(dt, tz=None))
        else:
            raise TypeError("End time must be a datetime object or a string.")

    @property
    def dt(self) -> DateTime:
        """The current real-world time for the model without timezone information."""
        return self._dt

    @dt.setter
    def dt(self, dt: DateTime) -> None:
        """Set the current real-world time."""
        if not isinstance(dt, DateTime):
            raise TypeError("dt must be a datetime object.")
        self._dt = dt

    @property
    def day(self) -> int:
        """Returns the current day for the model."""
        return self.dt.day

    @property
    def day_of_week(self) -> int:
        """Returns the number for the day of the week for the model."""
        return self.dt.day_of_week

    @property
    def day_of_year(self) -> int:
        """Returns the day of the year for the model."""
        return self.dt.day_of_year

    @property
    def days_in_month(self) -> int:
        """Get the total number of days of the month that this period falls on for the model."""
        return self.dt.days_in_month

    @property
    def hour(self) -> int:
        """Get the hour of the day component of the Period."""
        return self.dt.hour

    @property
    def minute(self) -> int:
        """Get minute of the hour component of the Period."""
        return self.dt.minute

    @property
    def month(self) -> int:
        """Return the month the current model's Period falls on."""
        return self.dt.month

    @property
    def quarter(self) -> int:
        """Return the quarter the current model's Period falls on."""
        return self.dt.quarter

    @property
    def second(self) -> int:
        """Get the second component of a model's Period."""
        return self.dt.second

    @property
    def is_leap_year(self) -> bool:
        """Return True if the period's year is in a leap year."""
        return self.dt.is_leap_year()

    @property
    def week_of_year(self) -> int:
        """Get the week of the year on the given Period."""
        return self.dt.week_of_year

    @property
    def week_of_month(self) -> int:
        """Get the week of the month on the given Period."""
        return self.dt.week_of_month

    @property
    def weekday(self) -> int:
        """Day of the week the period lies in, with Monday=0 and Sunday=6."""
        return self.dt.weekday()

    @property
    def year(self) -> int:
        """Return the year this Period falls on."""
        return self.dt.year

    def strftime(self, fmt: Optional[str] = None) -> str:
        """Returns a string representing the current time.

        Parameters:
            fmt:
                An explicit format string of datetime.
        """
        return (
            self.dt.strftime(self.fmt)
            if fmt is None
            else self.dt.strftime(fmt)
        )



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""
模型的基本模块。
Basic implementation of the model's module.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Iterator,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    cast,
)

from loguru import logger

from abses._bases.bases import _Notice
from abses._bases.objects import _BaseObj
from abses._bases.states import _States
from abses.tools.func import iter_func

if TYPE_CHECKING:
    from abses.main import MainModel

try:
    from typing import TypeAlias
except ImportError:
    from typing_extensions import TypeAlias

HowCreation: TypeAlias = Literal["from_resolution", "from_file", "copy_layer"]


class Module(_BaseObj):
    """Basic module for the model."""

    def __init__(
        self,
        model: MainModel[Any, Any],
        name: Optional[str] = None,
    ):
        _BaseObj.__init__(self, model, observer=True, name=name)
        self._open: bool = True

    def __repr__(self) -> str:
        flag = "open" if self.opening else "closed"
        return f"<{self.name}: {flag}>"

    @property
    def opening(self) -> bool:
        """If the module is open."""
        return self._open

    @opening.setter
    def opening(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise TypeError(f"Only accept boolean, got {type(value)}.")
        if self._open is not value:
            logger.info("%s switch 'open' to %s.", self.name, value)
        self._open = value

    def initialize(self):
        """
        Initialization before handle parameters.
        """

    def setup(self):
        """
        Initialization after handle parameters.
        """

    def step(self):
        """
        Called every time step.
        """

    def end(self):
        """
        Called at the end of the simulation.
        """


ModuleType = TypeVar("ModuleType", bound=Module)


class _ModuleFactory(Generic[ModuleType]):
    """To create a module."""

    methods: List[str] = []
    default_cls: Type[Module] = Module

    def __init__(self, father) -> None:
        self.father: CompositeModule = father
        self.modules: Dict[str, ModuleType] = {}

    def __str__(self) -> str:
        return f"{self.father}: {list(self.modules.keys())}"

    def __iter__(self) -> Iterator[Module]:
        return iter(self.modules.values())

    def __getitem__(self, name: str) -> Module:
        return self.modules[name]

    def __contains__(self, name: str | ModuleType) -> bool:
        return name in self.modules or name in self.modules.values()

    def __len__(self) -> int:
        return len(self.modules)

    def _check_cls(
        self, module_cls: Optional[Type[ModuleType]]
    ) -> Type[ModuleType]:
        """Check if the provided class is a valid module class."""
        if module_cls is None:
            return cast(Type[ModuleType], self.default_cls)
        if not issubclass(module_cls, self.default_cls):
            raise TypeError(
                f"'{module_cls}' not a subclass of {self.default_cls}."
            )
        return cast(Type[ModuleType], module_cls)

    @property
    def is_empty(self) -> bool:
        """If the factory is empty."""
        return len(self.modules.keys()) == 0

    def _check_name(self, name: str) -> None:
        """Check if the name is valid."""
        if name in self.modules:
            raise ValueError(f"Name '{name}' already exists in the model.")

    def new(
        self,
        how: Optional[HowCreation] = None,
        module_class: Optional[Type[ModuleType]] = None,
        **kwargs,
    ) -> ModuleType:
        """Create a module and attach it to the model.

        Parameters:
            module_class:
                The class of the module to be created.
                Must be a subclass of Module.
                If not given, the default module will be created.
            how:
                The method to create the module.
                If not given, the module will be created by its __init__ method.
            **kwargs:
                The parameters to initialize the module.

        Raises:
            TypeError:
                If the module class is not a subclass of Module.
            ValueError:
                If the creating method is not valid.

        Returns:
            The created module.
        """
        module_cls = self._check_cls(module_cls=module_class)
        if not how:
            module = module_cls(model=self.father.model, **kwargs)
        elif hasattr(self, how):
            module = getattr(self, how)(
                module_cls=module_cls, model=self.father.model, **kwargs
            )
        else:
            raise ValueError(
                f"{how} is not a valid method for creating module."
                f"Choose from {self.methods} for {self.default_cls}"
            )
        # register as module
        self._check_name(module.name)
        self.modules[module.name] = module
        self.father.attach(module)
        logger.info(f"{str(self.father)} created module {module.name}.")
        return module


# Composite
class CompositeModule(Module, _States, _Notice):
    """基本的组合模块，可以创建次级模块"""

    def __init__(
        self,
        model: MainModel[Any, Any],
        name: Optional[str] = None,
    ) -> None:
        Module.__init__(self, model, name=name)
        _States.__init__(self)
        _Notice.__init__(self)
        self._modules: _ModuleFactory = _ModuleFactory(self)

    @property
    def modules(self) -> _ModuleFactory:
        """All attached sub-modules."""
        return self._modules

    @property
    def opening(self) -> bool:
        return self._open

    @opening.setter
    def opening(self, value: bool) -> None:
        for module in self.modules:
            module.opening = value
        self._open = value

    @iter_func("modules")
    def initialize(self):
        return super().initialize()

    @iter_func("modules")
    def setup(self):
        return super().setup()

    @iter_func("modules")
    def step(self):
        return super().step()

    @iter_func("modules")
    def end(self):
        return super().end()

    def create_module(self, module_cls, how=None, **kwargs):
        """Create a module."""
        return self.modules.new(how=how, module_class=module_cls, **kwargs)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""Dynamic variables
"""

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Callable, List

if TYPE_CHECKING:
    from ..time import TimeDriver
    from .objects import _BaseObj


class _DynamicVariable:
    """Time dependent variable

    A time dependent function will take the model time driver as
    an input and return its value. The function can also take other
    variables as inputs. The function can be defined as a static
    method of a class or a function.
    """

    def __init__(
        self, name: str, obj: _BaseObj, data: Any, function: Callable, **kwargs
    ) -> None:
        self._name: str = name
        self._obj: _BaseObj = obj
        self._data: Any = data
        self._function: Callable = function
        self._cached_data: Any = None
        self.attrs = kwargs
        self.now()

    def __str__(self) -> str:
        return f"<{self.name}: {type(self.now())}>"

    def __repr__(self) -> str:
        return str(self)

    @property
    def name(self) -> str:
        """Get the name of the variable

        Returns
        -------
        name: str
        """
        return self._name

    @property
    def obj(self) -> _BaseObj:
        """Returns a base object instance

        Returns:
            obj:
                _BaseObj
        """
        return self._obj

    @obj.setter
    def obj(self, obj: _BaseObj):
        if not isinstance(obj, _BaseObj):
            raise TypeError("Only accept observer object")
        self._obj = obj

    @property
    def data(self) -> Any:
        """Returns unused data

        Returns:
            data:
                Any
        """
        return self._data

    @property
    def function(self) -> Callable:
        """Get the function that calculates the variable

        Returns:
            function:
                Callable
        """
        return self._function

    @property
    def time(self) -> TimeDriver:
        """Get the model time driver

        Returns:
            time:
                abses.time.TimeDriver
        """
        return self.obj.time

    def get_required_attributes(self, function: Callable) -> List[str]:
        """Get the function required attributes

        Returns:
            required_attributes:
                list[str]
        """
        # Get the source code of the function
        source_code = inspect.getsource(function)
        return [
            attr
            for attr in ["data", "obj", "time", "name"]
            if attr in source_code
        ]

    def now(self) -> Any:
        """Return the dynamic variable function's output

        Returns:
            The dynamic data value now.
        """
        required_attrs = self.get_required_attributes(self.function)
        args = {attr: getattr(self, attr) for attr in required_attrs}
        result = self.function(**args)
        self._cached_data = result
        return result

    @property
    def cache(self) -> Any:
        """Return the dynamic variable's cache"""
        return self._cached_data
